# $Id$

package CherryPick::LicenseManager;

use strict;
use base qw( MT::ErrorHandler );
use vars qw($DAY_SECONDS $LICENSE_FREE);
$DAY_SECONDS = 60 * 60 * 24;
$LICENSE_FREE = 0;

use MT;
use MT::I18N;
use Digest::MD5 qw(md5_hex);


sub new {
    my $class = shift;
    my ($params) = @_;
    my $self = $class->SUPER::new(@_);
    
    $self->{trial_days} = '14';
    $self->{confirm_interval_days} = '7';
    $self->{plugindata_key} = 'license';
    $self->{l10n_class} = 'CherryPick::LicenseManager::L10N';

    $self->{setup_time_name} = 'setup_time';
    $self->{licensekey_name} = 'licensekey';
    $self->{email_name} = 'email';
    $self->{last_message_name} = 'last_message';
    $self->{confirmed_time_name} = 'confirmed_time';
    $self->{confirmation_message_name} = 'confirmation_message';
    
    foreach my $name (keys %$params) { $self->{$name} = $params->{$name} };
    
    if(!$self->{trial} && $self->{trial_days}) {
        $self->{trial} = $self->{trial_days} * $DAY_SECONDS; # days to seconds.
    }
    if(!$self->{confirm_interval} && $self->{confirm_interval_days}) {
        $self->{confirm_interval} = $self->{confirm_interval_days} * $DAY_SECONDS; # days to seconds.
    }
    
    foreach my $member (qw(plugin setup_time_name licensekey_name email_name last_message_name confirmed_time_name confirmation_message_name trial plugindata_key)) {
        return $class->error($self->translate('CherryPick::LicenseManager new object requires [_1] member.', $member))
            unless $self->{$member};
    }
    
    $self->{plugin_key} ||= ($self->{plugin}->key || $self->{plugin}->name);
    
    my $time = time;
    
    unless($self->setup_time) {
        $self->setup_time($time);
        $self->save;
    }
    
    return $self if $LICENSE_FREE;
    
    if($self->{confirm_interval} && $self->licensekey) {
        my $confirmed_time = $self->confirmed_time || $self->setup_time;
        if($time - $confirmed_time > $self->{confirm_interval}) {
            $self->confirm_license;
        }
    }
    
    return $self;
}

sub component {
    my $self = shift;
    $self->{plugin} || MT->component('core');
}

sub translate {
    my $self = shift;
    my $c = $self->component;
    my $lang = MT->current_language || MT->config->DefaultLanguage;
    
    my $handles = MT->request('l10n_handle') || {};
    my $ex_handle = $handles->{$c->id};
    my $ex_class = $c->l10n_class;
    $c->l10n_class($self->{l10n_class});
    delete $handles->{$c->id};
    
    my $result = $c->translate(@_);
    
    delete $handles->{$c->id};
    $handles->{$c->id} = $ex_handle if $ex_handle;
    $c->l10n_class($ex_class);
    
    return $result;
}

sub plugin_data {
    my $self = shift;
    return $self->{plugin_data} if $self->{plugin_data};
    
    my $pd = MT->model('plugindata')->load({
        plugin => $self->{plugin_key},
        key => $self->{plugindata_key},
    });
    
    unless($pd) {
        $pd = MT->model('plugindata')->new || return $self->error(MT->model('plugindata')->errstr);
        $pd->plugin($self->{plugin_key});
        $pd->key($self->{plugindata_key});
        $pd->data({})
    }
    
    $self->{plugin_data} = $pd;
}

sub save {
    my $self = shift;
    
    my $pd = $self->plugin_data || return;
    $pd->save || return $self->error($pd->errstr);
}

sub data_value {
    my $self = shift;
    my ($name, $value) = @_;
    return unless $name;
    
    my $plugin_data = $self->plugin_data || return;
    my $data = $plugin_data->data || {};
    
    if(defined($value)) {
        $data->{$name} = $value;
        $plugin_data->data($data);
    }
    return '' unless defined($name) && defined($data->{$name});
    
    $data->{$name};
}

sub setup_time {
    my $self = shift;
    my ($time) = @_;
    
    $self->data_value($self->{setup_time_name}, $time);
}

sub confirmed_time {
    my $self = shift;
    my ($time) = @_;
    
    $self->data_value($self->{confirmed_time}, $time);
}

sub licensekey {
    my $self = shift;
    my ($licensekey) = @_;
    
    $self->data_value($self->{licensekey_name}, $licensekey);
}

sub last_message {
    my $self = shift;
    my ($message) = @_;
    
    my $value = $self->data_value($self->{last_message_name}, $message);
    
    $self->save;

    $value;
}

sub confirmation_message {
    my $self = shift;
    my ($message) = @_;
    
    my $value = $self->data_value($self->{confirmation_message_name}, $message);
    
    $self->save;
    
    $value;
}

sub email {
    my $self = shift;
    my ($email) = @_;
    
    $self->data_value($self->{email_name}, $email);
}

sub is_expired {
    my $self = shift;
    
    my $setup_time = $self->setup_time;
    my $licensekey = $self->licensekey;
    return 0 if $LICENSE_FREE;
    return !$licensekey && time > $setup_time + $self->{trial}? 1: 0;
}

sub warning {
    my $self = shift;
    
    $self->confirmation_message;
}

sub remainder {
    my $self = shift;
    
    my $setup_time = $self->setup_time;
    my $remainder = $setup_time + $self->{trial} - time;
    
    $remainder = 0 if $remainder < 0;
    $remainder;
}

sub remainder_days {
    my $self = shift;
    
    my $remainder = $self->remainder;
    $remainder? int(($self->remainder + $DAY_SECONDS / 2) / $DAY_SECONDS): 0; #seconds to days.
}

sub is_registered {
    my $self = shift;
    my $licensekey = $self->licensekey;
    return 1 if $LICENSE_FREE;
    return $licensekey? 1: 0;
}

sub post_license {
    my $self = shift;
    my ($url, $post) = @_;
    
    my $app = MT->instance;
    if ( MT->version_number >= 5 ) {
        require MT::Website;
    } else {
        require MT::Blog;
    }
    my $systemLocation;
    unless ( ref($app) eq 'MT::App::CMS' ) {
        my $blog;
        if ( MT->version_number >= 5 ) {
            $blog = MT::Website->load({}, {limit => 1});
        } else {
            $blog = MT::Blog->load({}, {limit => 1});
        }
        $systemLocation = $blog->site_url;
        $systemLocation = $1 if $systemLocation =~ m!(https?://[^/]+?/).+!;
        $systemLocation =~ s!/$!! if $systemLocation =~ m!/$!;
        $systemLocation .= mt_uri($app);
    }
    else {
        $systemLocation = $app->base .$app->mt_uri;
    }
    $post->{pluginName} = $self->{plugin_key};
    $post->{systemLocation} = $systemLocation;
    $post->{hint} ||= '';
    $post->{lang} ||= (MT->current_language || MT->config->DefaultLanguage);
    
    my $ua = MT->instance->new_ua;
    $ua->post($url, $post);
}

sub mt_path {
    my $app = shift;
    require MT::App;
    MT::App::mt_path($app);
}

sub mt_uri {
    my $app = shift;
    mt_path($app) . $app->config->AdminScript;
}

sub normalize_response {
    my $self = shift;
    my ($message) = shift;
    
    if(MT->version_number >= 5) {
        utf8::decode($message);
    } else {
        my $charset = MT->instance->config('PublishCharset') || 'UTF-8';
        $message = MT::I18N::convert_text('UTF-8', $charset, $message) unless $charset =~ /UTF-?8/i;
    }
    
    $message;
}

sub register_license {
    my $self = shift;
    my ($param) = @_;
    my $post = {};
    
    $self->last_message('');
    
    unless($param->{licensekey}) {
        $self->last_message($self->translate('The license key is required.'));
        return 'error';
    }
    
    unless($param->{email}) {
        $self->last_message($self->translate('The email is required.'));
        return 'error';
    }
    
    my $url = $self->{register_url};
    unless($url) {
        $self->last_message($self->translate('The register_url is not defined.'));
        return 'error';
    }

    my ($result, $message);
    if ( !$self->{local_activation} ) {
        my $response = $self->post_license($url, {
            licenseKey      => $param->{licensekey},
            mailAddress     => $param->{email},
            hint            => $param->{hint},
        });
        
        if(!$response->is_success || !$response->headers->header('X-Result')) {
            $self->last_message($self->translate('Failure to connect to the licensing server for registration.'));
            return 'error';
        }

        $result = $response->headers->header('X-Result');
        $message = $self->normalize_response($response->headers->header('X-Message'));
    } else {
        # Get local hash 
        require File::Spec;
        my $plugin = $self->{plugin};
        my $path = File::Spec->catfile($plugin->{full_path}, 'tmp', 'license.md5');
        unless ( -f $path ) {
            $result = 'error';
            $message = $self->translate('Cannot activate your license.');
        } else {
            open my $fh, "<$path";
            my $correct = '';
            while ( <$fh> ) {
                $correct .= $_;
            }
            close $fh;
            my $hash = md5_hex($param->{licensekey});
            if ( $correct && $correct == $hash ) {
                $result = 'success';
                $message = $self->translate('Thank you for your purchase. The license key has activated.');
            } else {
                $result = 'error';
                $message = $self->translate('Cannot activate your license.');
            }
        }
    }
    
    if($result ne 'error') {
        $self->confirmation_message('');
        $self->licensekey($param->{licensekey});
        $self->email($param->{email});
        $self->save;
    }
    
    $self->last_message($message);

    return $result;
}

sub confirm_license {
    my $self = shift;
    my ($param) = @_;
    my $post = {};

    # Skip once
    return 'complete';
    
    my $url = $self->{confirm_url} || return 'complete';
    
    $self->confirmation_message('');
    
    my $response = $self->post_license($url, {
        licenseKey      => $self->licensekey,
    });
    
    $self->confirmed_time(time);
    $self->save;
    
    if(!$response->is_success || !$response->headers->header('X-Result')) {
        MT::log($self->translate('Failure to connect to the licensing server for confirmation.'));
        return 'warning';
    }
    
    my $result = $response->headers->header('X-Result');
    my $message = $self->normalize_response($response->headers->header('X-Message'));
    
    if($result eq 'error') {
        $self->confirmation_message($message);
        
        $self->licensekey('');
        $self->email('');
        $self->save;
    } elsif($result ne 'complete') {
        MT::log($message) if $message;
    }
    
    return $result;
}

1;