package Uploader::App;

use vars qw($TRANSFER_TIMEOUT $STEPS);
$TRANSFER_TIMEOUT = 120;
$STEPS = qq{[["copy"],["replace"],["custom"],["transfer"],["delete"]]}; #qq{[["copy"],["replace"],["custom"],["transfer"],["delete"]]}

use strict;

use base 'MT::App::CMS';

use File::Spec;
use LWP::UserAgent;
use MT;
use MT::Util;
use MT::Blog;
use MT::I18N;
use JSON;
use Time::HiRes;

use vars qw($MAX_TIME);
sub BEGIN {
    $MAX_TIME = 1200;
}

sub init {
    my $app = shift;

    $app->SUPER::init(@_) or return;
    $app->add_methods(
            download => \&download,
            transfer => \&transfer,
            conftransfer => \&conftransfer,
            run_actions => \&run_actions,
            get_file => \&get_file,
            test => \&test,
    );
    $app->{default_mode} = 'download';
    $app->{requires_login} = 1;
    $app->{is_admin} = 1;
    $app->{plugin_template_path} = $app->template_dir()
        if MT->version_number >= 5;

    $app;
}

sub template_dir {
    my $app = shift;
    my $plugin = plugin();
    my $mt_dir = $app->mt_dir;
    my $envelope = $plugin->envelope;
    my $path = File::Spec->catdir($mt_dir, $envelope, 'tmpl');
    $path;
}

sub plugin {
    MT::Plugin::Uploader->instance;
}

sub build_page {
    my $app = shift;
    my ($template, $param) = @_;


    my $plugin = $app->plugin;
    if ($plugin) {
        my $path = $app->static_path;
        $path .= '/' unless $path =~ m!/$!;
        $path .= $app->plugin->envelope . "/";
        $path = $app->base . $path if $path =~ m!^/!;
        $_[1]->{plugin_static_uri} = $path;
    }

    $param->{script_url} = $app->config->CGIPath .$app->plugin->envelope  . '/uploader.cgi';
    if(MT->version_number < 5) {
        $template = "mt4/$template" ;
        $app->MT::App::build_page($template, $param);
    } elsif (MT->version_number >= 5) {
        $template = "mt5/$template" ;
        $app->MT::App::build_page($template, $param);
    } else {
        $app->MT::App::CMS::build_page($template, $param);
    }
}

sub json_response {
    my $app = shift;
    $app->print(' JSON:' . objToJson($app->response));
}

sub response {
    my $self = shift || MT->instance;
    return unless ref $self;
    if (!$self->{response}) {
        $self->{response} = { steps => [], progress => [], error => undef };
    }
    $self->{response};
}

sub progress {
    my $app = shift;
    my ($msg, $make_id) = @_;
    $msg =~ s/^\s+//gs;
    $msg =~ s/\s+$//gs;
    $msg =~ s/\s+/ /gs;
    if ($make_id) {
        require MT::Util;
        my $id = MT::Util::dirify($make_id);
        $msg = qq{#$id $msg};
    }
    my $print = ($app->version_number >= 5) ? 'print_encode' : 'print';
    $app->$print( $msg ."\n" );
    $app->json_response();
    exit 1;
}

sub error {
    my $app = shift;
    my ($msg) = @_;
    $app->SUPER::error(@_);
    return unless ref $app;
    $app->response->{error} = $msg;
    return;
}

sub check_authorization {
    my $app = shift;
    my $user = $app->user || return 0;
    return 1 if $user->is_superuser;
    my $perms = $app->user->blog_perm($app->param('blog_id')) || return 0;
    return ( $perms->can_administer_blog || $perms->can_uploader );
}

sub test {
    my $app = shift;
    $app->error('test');
}

sub download {
    my $app = shift;

    return $app->error($app->translate('Permission denied.'))
        unless $app->check_authorization;

    my %param;
    $param{downloading} = 1;
    $param{goback} = $app->{goback} || 'history.back()';
    $param{steps} = qq{[["copy"],["replace"],["custom"],["compress"],["delete"]]};
    $param{blog_id} = scalar $app->param('blog_id');

    if ( $app->config('UploaderUniqueTempDir') ) {
        my $session = Time::HiRes::time;
        $param{session_key} = $session;
    }

    return $app->build_page('upload_runner.tmpl', \%param);
}

sub reservation_status {
    my ( $app, $param ) = @_;
    my $blog = $app->blog;

    my ( %system, %config );
    plugin->load_config( \%system, 'system' );
    plugin->load_config( \%config, 'blog:' . $blog->id );

    my $message;
    if ( $config{reservation} ) {
        if ( $config{reservation_retried} == 0 ) {

            # 予約済みだが未実行
            $message = 'Transfering reserved at [_1].';
        } else {

            # 実行済み
            if ( $config{last_reserve_transfered_epoch} ) {

                # 完了済み
                if ( $config{reservation_retried} < $system{reservation_retry} ) {

                    # 成功
                    $message = 'Reserved transfering succeeded at [_2]';
                } else {

                    # 失敗
                    $message = 'Reserved transfering failed after [_4] time(s) retry at [_2].';
                }
            } else {

                # 実行中・再試行中
                $message = 'Reserved transfering running: [_3]/[_4] time(s) retried.';
            }
        }
    } else {

        # 予約されていない
        $message = 'Transfering not reserved.';
    }

    plugin->translate($message,
        MT::Util::format_ts('%Y-%m-%d %H:%M:%S', MT::Util::epoch2ts($blog, $config{reservation_epoch}), $blog),
        MT::Util::format_ts('%Y-%m-%d %H:%M:%S', MT::Util::epoch2ts($blog, $config{last_reserve_transfered_epoch}), $blog),
        $config{reservation_retried},
        $system{reservation_retry},
    );
}

sub display_reserve_form {
    my ( $app, $param ) = @_;
    my $q = $app->param;
    my $blog = $app->blog;

    my %config;
    plugin->load_config(\%config, 'blog:' . $blog->id);

    my $current_ts = MT::Util::epoch2ts($blog, time) || 0;
    $param->{current_datetime} = MT::Util::format_ts('%Y-%m-%d %H:%M:%S', $current_ts, $blog) || '';
    $param->{reservation_status} = reservation_status($app, $param);
    $param->{batch_on} = $config{'do_run-periodical-tasks'};

    plugin->load_tmpl('reserve.tmpl', $param);
}

sub reserve {
    my ( $app ) = @_;
    my $q = $app->param;
    my $blog = $app->blog
        or $app->return_to_dashboard( redirect => 1 );
    my $scope = 'blog:' . $blog->id;

    my %param;
    my %config;

    plugin->load_config(\%config, $scope);
    require MT::Util;
    my $epoch = $config{reservation_epoch};
    my $ts = $epoch ? MT::Util::epoch2ts($blog, $epoch) : 0;

    $param{reservation} = $config{reservation} && $ts ? 1 : 0;
    $param{reservation_ts_date} = $ts ? MT::Util::format_ts('%Y-%m-%d', $ts, $blog) : '';
    $param{reservation_ts_time} = $ts ? MT::Util::format_ts('%H:%M:%S', $ts, $blog) : '';

    $param{saved} = $q->param('saved') || 0;
    $param{canceled} = $q->param('canceled') || 0;

    display_reserve_form($app, \%param);
}

sub save_reserve {
    my ( $app ) = @_;
    my $q = $app->param;
    my $blog = $app->blog;
    my $scope = 'blog:' . $blog->id;

    my $reservation = $q->param('reservation');
    my $reservation_epoch = 0;

    if ( $reservation ) {
        my ( $ts_date, $ts_time ) = map { $q->param('reserve_on_' . $_) || ''; } qw/date time/;
        my $ts = join('', $ts_date, $ts_time);
        $ts =~ s/\D//g;
        $ts .= 0 x (14 - length($ts)) if length($ts) < 14;


        require MT::Util;
        $reservation_epoch = MT::Util::ts2epoch($blog, $ts);

        # Error if past time
        if ( !$MT::DebugMode && $reservation_epoch < time ) {
            my %param = $app->param_hash;
            $param{invalid_date} = 1;
            $param{reservation_ts_date} = $ts_date;
            $param{reservation_ts_time} = $ts_time;
            $param{$_} = $q->param($_) foreach qw/reserve_on_date reserve_on_time/;

            return display_reserve_form($app, \%param);
        }
    }

    # Save config
    my %config;
    plugin->load_config(\%config, $scope);
    $config{reservation} = $reservation;
    $config{reservation_epoch} = $reservation_epoch;
    $config{reservation_retried} = 0;
    $config{last_reserve_transfered_epoch} = 0;
    plugin->save_config(\%config, $scope);

    $app->redirect(
        $app->uri(
            mode => 'uploader_reserve',
            args => {
                blog_id => $blog->id,
                saved => $reservation ? 1 : 0,
                canceled => $reservation ? 0 : 1,
            },
        )
    );
}

sub conftransfer {
    my $app = shift;

    return $app->error($app->translate('Permission denied.'))
        unless $app->check_authorization;

    my %param;
    $param{downloading} = 0;
    $param{goback} = $app->{goback} || 'history.back()';
    $param{steps} = $STEPS;
    $param{blog_id} = scalar $app->param('blog_id');

    return $app->build_page('confirm.tmpl', \%param);
}

sub transfer {
    my $app = shift;

    return $app->error($app->translate('Permission denied.'))
        unless $app->check_authorization;

    my %param;
    $param{downloading} = 0;
    $param{goback} = $app->{goback} || 'history.back()';
    $param{steps} = $STEPS;
    $param{blog_id} = scalar $app->param('blog_id');
    $param{warning_transfer_timeout} = $app->config('UploaderWarningTransferTimeout');

    if ( $app->config('UploaderUniqueTempDir') ) {
        my $session = Time::HiRes::time;
        $param{session_key} = $session;
    }

    return $app->build_page('upload_runner.tmpl', \%param);
}

sub get_configs {
    my $app = shift;
    my $blog_id = scalar $app->param('blog_id');
    return $app->error($app->translate('blog_id is required')) unless $blog_id;

    my $mt_path = MT->instance->mt_dir;
    my $plugin_dir = File::Spec->catdir($mt_path, $app->plugin->envelope);

    my $config = $app->plugin->get_config_hash();
    my $blog_config = $app->plugin->get_config_hash('blog:'.$blog_id);

    my $base_dir = $config->{temp_dir} || File::Spec->catdir($plugin_dir, '/tmp/');

    return ($config, $blog_config, $mt_path, $blog_id, $base_dir);
}

sub run_actions {
    my $app = shift;


    $| = 1;
    $app->{no_print_body} = 1;
    $app->send_http_header('text/plain');

    my ($config, $blog_config, $mt_path, $blog_id, $base_dir) = $app->get_configs;
    $config->{$_} ||= $blog_config->{$_} foreach keys %$blog_config;

    my $temp_dir = File::Spec->catdir($base_dir, $blog_id);

    if ( $app->config('UploaderUniqueTempDir') ) {
        my $session = $app->param('session') || '';
        $temp_dir .= "-$session" if $session;
    }

    $temp_dir = qq("$&") if $temp_dir =~ /\s/;

    my $blog = MT::Blog->load($blog_id);
    return $app->error($app->translate_json('Blog is not valid')) unless $blog;
    my $blog_dir = $blog->site_path;
    $blog_dir = qq("$&") if $blog_dir =~ /\s/;

    my $replace_table = $blog_config->{replace_table};

    my $compress_format = $blog_config->{compress_format} || 'zip';
    my $compress_file = File::Spec->catdir($base_dir, "download-blog-$blog_id");
    $compress_file = qq("$compress_file") if $compress_file =~ /\s/;

    my $transfer = $blog_config->{transfer} || 'ftp';

    my $exclude_dirs = $blog_config->{exclude_dirs};

    my $ftp_server = $blog_config->{ftp_server};
    my $ftp_user = $blog_config->{ftp_user};
    my $ftp_password = $blog_config->{ftp_password};
    my $ftp_remote_dir = $blog_config->{ftp_remote_dir};

    my $sync_dir = $blog_config->{sync_dir};
    $sync_dir = qq("$sync_dir") if $sync_dir =~ /\s/;

    my $custom_actions = $blog_config->{custom_actions};

    my $steps = scalar $app->param('steps');
    $steps = jsonToObj($steps);

    my $start = time;
    my @steps = ( @$steps );
    my $step;

    my $version = $app->version_number;

    eval {
        while($step = shift @steps) {
            $app->response->{steps} = \@steps if $version >= 5;
            if($step->[0] eq 'copy') {
                $app->step_copy($config, $blog_dir, $temp_dir, $exclude_dirs) || last;
                $app->progress($app->translate_json('Copy blog contents to temporary.'));
            } elsif($step->[0] eq 'replace') {
                $app->step_replace($config, $temp_dir, $replace_table) || last;
                $app->progress($app->translate_json('Replace the temporary contents with table.'));
            } elsif($step->[0] eq 'compress') {
                if($compress_format eq 'zip') {
                    $app->step_compress_zip($config, $temp_dir, $compress_file) || last;
                } else {
                    $app->step_compress_targz($config, $temp_dir, $compress_file) || last;
                }
                $app->progress($app->translate_json('Compress the temporary contents.'));
            } elsif($step->[0] eq 'transfer') {
                if($transfer eq 'ftp') {
                    my $res = $app->step_transfer_ftp($config, $temp_dir, $ftp_server, $ftp_user, $ftp_password, $ftp_remote_dir);

                    my $msg = 'Transfer and synchronize the temporary contents via FTP.';
                    if ( $res && $res !~ /^1$/ ) {
                        shift @steps;
                        $msg = 'Transfering and synchronizing the temporary contents via FTP was time out [_1] seconds.';
                        $app->response->{steps} = \@steps;
                    }

                    my $timeout = MT->instance->config('UploaderTransferTimeout');
                    $timeout = $TRANSFER_TIMEOUT unless defined $timeout;

                    $app->progress($app->translate_json($msg, $timeout));

                    die $res if $res && $res !~ /^1$/;

                } else {
                    $app->step_transfer_local($config, $temp_dir, $sync_dir) || last;
                    $app->progress($app->translate_json('Transfer and synchronize the temporary contents to the directory.'));
                }
            } elsif($step->[0] eq 'delete') {
                $app->step_delete($config, $temp_dir) || last;
                $app->progress($app->translate_json('Delete the temporary contents.', $temp_dir));
            } elsif($step->[0] eq 'custom') {
                $app->step_custom($config, $custom_actions, $base_dir, $temp_dir, $blog_id) || last;
                $app->progress($app->translate_json('Process custom actions.'));
            } else {
                $app->error($app->translate('[_1] is bad step', $step));
                last;
            }
        }
        last if time > $start + $MAX_TIME;
    };
    if($@) {
        unshift @steps, $step if $step;
        $app->response->{error} = $@
            unless $@ =~ /timeout/;
    }
    if($app->errstr) {
        unshift @steps, $step;
        $app->response->{error} = $app->errstr;
    }

    if(@steps) {
        $app->response->{steps} = \@steps;
    }
    $app->json_response;
}

sub step_copy {
    my $app = shift;
    my ($config, $blog_dir, $temp_dir, $exclude_dirs) = @_;
    my $result;
    my $script;
    my @exclude_dir_list;
    my $exclude_command;

    $app->step_delete($config, $temp_dir);

    @exclude_dir_list = grep { $_ } map { s/(^\s+)|(\s+$)//g; $_ } split(/\r?\n/, $exclude_dirs);

    my $unique_temp_dir = MT->instance->config('UploaderUniqueTempDir');
    my $copy_with_rsync = MT->instance->config('UploaderCopyWithRsync');
    if ( !$unique_temp_dir && $copy_with_rsync ) {
        my $rsync = MT->instance->config('UploaderRsyncPath') || 'rsync';
        my $excludes = join ' ', map { s/'/\\'/; "--exclude '$_'" } @exclude_dir_list;
        my $options = MT->instance->config('UploaderRsyncCopyOption') || '-auv --delete';
        $script = "$rsync $options '$blog_dir/' '$temp_dir' $excludes 2>&1";
    } else {
        $script = "cp -a '$blog_dir' '$temp_dir' 2>&1";
    }

    MT::log($script) if $MT::DebugMode;
    $result = `$script`;
    MT::log($result) if $MT::DebugMode;
    return $app->error($result) if $? != 0 ;

    if ( @exclude_dir_list ) {
        foreach ( @exclude_dir_list ) {
            my $del_path = File::Spec->catdir($temp_dir, $_);
            next unless defined $del_path && -e $del_path;
            $script = "rm -rf '$del_path'";
            $result = `$script`;
            return $app->error($result) if $? != 0 ;
        }
    }
    1;
}

sub step_replace {
    my $app = shift;
    my ($config, $temp_dir, $table) = @_;
    my $result;
    my $script;

    if( $config->{replace_script} ){
        $script = "$config->{replace_script} 2>&1";
        $result = `$script`;
        return $app->error($result) if $? != 0 ;
    }else{
        $table =~ s/\r//g;
        foreach (split(/\n/, $table)) {
            my ($cond, $exec) = split(/:/, $_, 2);
            $cond = "*.$1" if $cond =~ /^\*([^.].*?)$/;
            if($exec) {
                $cond = $cond? "-name '$cond'": '-type f';
                $script = "find '$temp_dir' $cond -print0 | xargs -0 sed -i -e '$exec' 2>&1";
                $result = `$script`;
                if ($? != 0) {
                    return $app->error($result) unless $result =~ /no input files/;
                }
            }
        }
    }

    1;
}

sub step_compress_zip {
    my $app = shift;
    my ($config, $temp_dir, $compress_file) = @_;
    my $result;
    my @scripts;

    if( $config->{zip_script} ){
        my $script = "$config->{zip_script} 2>&1";
        $result = `$script`;
    }else{
        @scripts = ("cd '$temp_dir';zip -r $compress_file.zip ./* 2>&1",
            "mv $compress_file.zip $compress_file 2>&1");
        foreach( @scripts ){
            $result = `$_`;
            return $app->error($result."<br/> command=$_ <br/> status=$?") if $? != 0 ;
        }
    }
    1;
}

sub step_compress_targz {
    my $app = shift;
    my ($config, $temp_dir, $compress_file) = @_;
    my $result;
    my @scripts;

    @scripts = (
        "cd '$temp_dir';tar cvzf $compress_file.tar.gz ./* 2>&1",
        "mv $compress_file.tar.gz $compress_file 2>&1");
    @scripts = ("$config->{tar_script} 2>&1") if $config->{tar_script} ;

    foreach( @scripts ){
        $result = `$_`;
        return $app->error($result) if $? != 0 ;
    }

    1;
}

sub step_transfer_ftp {
    my $app = shift;
    my ($config, $temp_dir, $ftp_server, $ftp_user, $ftp_password, $ftp_remote_dir) = @_;
    my $result;
    my $script;
    my $lftp = MT->instance->config('UploaderLftpPath') || 'lftp';
    my $options = $config->{lftp_mirror_options} || '-R -n -e';
    my $lftp_pre_command = $config->{lftp_pre_command} || '';
    $lftp_pre_command .= ' &&' if $lftp_pre_command;

    $script = "$lftp -c \"$lftp_pre_command open $ftp_server && user $ftp_user $ftp_password && lcd '$temp_dir' && mirror $options ./ '$ftp_remote_dir' \" 2>&1";
    $script = "$config->{ftp_script} 2>&1" if $config->{ftp_script} ;

    # FTP Proxy対応
    if ( MT->instance->config('UploaderEnableProxy') ) {
        my $use_proxy = defined($config->{use_proxy}) ? $config->{use_proxy} : '';
        my $proxy_type = defined($config->{proxy_type}) ? $config->{proxy_type} : '';
        my $proxy_host = defined($config->{http_proxy_host}) ? $config->{http_proxy_host} : '';
        my $proxy_port = defined($config->{http_proxy_port}) ? $config->{http_proxy_port} : '';
        my $proxy_user = defined($config->{http_proxy_user}) ? $config->{http_proxy_user} : '';
        my $proxy_pass = defined($config->{http_proxy_pass}) ? $config->{http_proxy_pass} : '';

        if ( $use_proxy ) {
            if ( $proxy_type eq 'ftp' ) {
                $script = $lftp . ' -c "' . $lftp_pre_command . ' set ftp:proxy ftp://' .$proxy_user .':' .$proxy_pass .'@' .$proxy_host .':' .$proxy_port;
                $script .= " && open $ftp_server";
            }
            else {
                $script = $lftp . ' -c "' . $lftp_pre_command . ' set hftp:proxy http://' .$proxy_user .':' .$proxy_pass .'@' .$proxy_host .':' .$proxy_port;
                $script .= " && open hftp://$ftp_server";
            }
            $script .= " && user $ftp_user $ftp_password";
            $script .= " && lcd '$temp_dir'";
            $script .= " && mirror $options ./ '$ftp_remote_dir' \" 2>&1";
        }
    }

    MT::log($script) if $MT::DebugMode;

    $@ = undef;
    my $timeout = MT->instance->config('UploaderTransferTimeout');
    $timeout = $TRANSFER_TIMEOUT unless defined $timeout;

    eval {
        local $SIG{ALRM} = sub { die "timed out $timeout seconds." };
        alarm $timeout;
        $result = `$script`;
        alarm 0;
    };
    return $@ if $@;
    return $app->error('Failed to Transfer via FTP. '. $result) if $? != 0 ;

    1;
}

sub step_transfer_local {
    my $app = shift;
    my ($config, $temp_dir, $sync_dir) = @_;
    my $result;
    my $script;
    my $rsync = MT->instance->config('UploaderRsyncPath') || 'rsync';

    $script = "$rsync -auv --delete '$temp_dir/' '$sync_dir' 2>&1";
    if ( $config->{rsync_script} ) {
        $script = "$config->{rsync_script} '$temp_dir/' '$sync_dir' 2>&1";
    }
    MT::log($script) if $MT::DebugMode;
    $result = `$script`;
    return $app->error($result) if $? != 0 ;

    1;
}

sub step_delete {
    my $app = shift;
    my ($config, $temp_dir) = @_;
    my $result;
    my $script;

    my $unique_temp_dir = MT->instance->config('UploaderUniqueTempDir');
    return 1 if $unique_temp_dir;

    my $copy_with_rsync = MT->instance->config('UploaderCopyWithRsync');
    return 1 if $copy_with_rsync;

    $script = "rm -rf '$temp_dir' 2>&1";
    $script = "$config->{delete_script} 2>&1" if $config->{delete_script} ;

    my @parts;
    @parts = split(/\//, $temp_dir);
    @parts = split(/\\/, $temp_dir) unless @parts;
    return $app->error($app->translate_json('About to delete shallow directory [_1]', $temp_dir)) if @parts < 3;

    MT::log($script) if $MT::DebugMode;
    $result = `$script`;
    return $app->error($result) if $? != 0 ;

    1;
}

sub step_custom {
    my $app = shift;
    my ($config, $custom_actions, $base_dir, $temp_dir, $blog_id) = @_;

    $custom_actions =~ s/\r//g;
    $custom_actions =~ s/\[base_dir\]/$base_dir/g;
    $custom_actions =~ s/\[temp_dir\]/$temp_dir/g;
    $custom_actions =~ s/\[blog_id\]/$blog_id/g;

    my @commands = split(/\n/, $custom_actions);

    foreach (@commands){
        my $result = `$_`;
        return $app->error($result. "<br/>command=$_ <br/>status=$?") if $? != 0 ;
    }

    1;
}

sub get_file {
    my $app = shift;

    return $app->error($app->translate('Permission denied.'))
        unless $app->check_authorization;

    my ($config, $blog_config, $mt_path, $blog_id, $base_dir) = $app->get_configs;

    my $compress_format = $blog_config->{compress_format};
    my $ext = $compress_format || 'zip';
    my $file = "download-blog-$blog_id";
    my $compress_file = File::Spec->catdir($base_dir, $file);

    my ($sec, $min, $hour, $mday, $mon, $year) = localtime();
    my $date = sprintf('%04d-%02d-%02d_%02d-%02d-%02d', $year + 1900, $mon + 1, $mday, $hour, $min, $sec);

    my $size = -s $compress_file;
    open F, "<$compress_file";
    binmode F;
    my $content;
    my $bytes = sysread F, $content, $size;
    close F;

    $| = 1;
    $app->{no_print_body} = 1;
    $app->set_header("Content-Disposition" => "attachment; filename=${file}_${date}.$ext");
    $app->set_header("Content-Length" => $size);
    $app->send_http_header('application/zip');

    $app->print($content);
}

sub translate {
    my $app = shift;
    $app->plugin->translate(@_);
}

sub translate_json {
    my $app = shift;
    my $trans = $app->plugin->translate(@_);
    $trans = MT::I18N::encode_text($trans, undef, 'utf-8');
    return MT::Util::escape_unicode($trans);
}

sub translate_templatized {
    my $app = shift;
    $app->plugin->translate_templatized(@_);
}

1;
