package Uploader::Task;

use vars qw($TRANSFER_TIMEOUT $STEPS);
$TRANSFER_TIMEOUT = 600;
$STEPS = qq{[["copy"],["replace"],["custom"],["transfer"],["delete"]]}; #qq{[["copy"],["replace"],["custom"],["transfer"],["delete"]]}

use strict;

use File::Spec;
use MT;
use MT::Util;
use MT::Blog;
use MT::I18N;
use Encode;
use Time::HiRes;

sub upload {
    my $task = shift;

    my $plugin = plugin();
    unless ( $plugin->is_enable ) {
        MT::log( $plugin->translate("Trial is expired.") );
        return 1;
    }

    my $blogs = sync_blogs();
    return unless @$blogs;

    foreach my $blog ( @$blogs ) {
        my ($config, $blog_config, $mt_path, $blog_id, $base_dir) = get_configs($blog);

        if ( $blog_config->{reservation} && !$blog_config->{last_reserve_transfered_epoch} ) {
            next if $blog_config->{reservation_epoch} > time;
            next unless increment_retry($plugin, $blog, $blog_config, $config);
            MT::log(
                $plugin->translate('Starting reserved transfer abount blog: [_1] (_2)',
                    $blog->name,
                    $blog_config->{reservation_retried},
                )
            );
        } else {
            next unless $blog_config->{'do_run-periodical-tasks'};
        }

        $config->{$_} ||= $blog_config->{$_} foreach keys %$blog_config;

        my $temp_dir = File::Spec->catdir($base_dir, $blog_id);

        if ( MT->instance->config('UploaderUniqueTempDir') ) {
            my $session = Time::HiRes::time;
            $temp_dir .= "-$session";

            # Cleanup
            my $expires = MT->instance->config('UploaderUniqueTempDirExpires');
            $expires = 60 * 60 * 24 unless defined $expires;

            if ( $expires ) {
                if ( opendir(DH, $base_dir) ) {
                    while ( my $d = readdir(DH) ) {
                        next if $d =~ /^\.+$/;
                        my $path = File::Spec->catdir($base_dir, $d);
                        next unless -d $path;
                        if ( $d =~ /^$blog_id-([0-9\.]+)$/ ) {
                            my $time = $1;
                            if ( $session - $time > $expires ) {
                                my $script = "rm -rf '$path' 2>&1";
                                MT::log($script) if $MT::DebugMode;
                                `$script`;
                            }
                        }
                    }
                    closedir DH;
                }
            }
        }

        $temp_dir = qq("$&") if $temp_dir =~ /\s/;

        return MT::log( $plugin->translate('Blog is not valid') ) unless $blog;
        my $blog_dir = $blog->site_path;
        $blog_dir = qq("$&") if $blog_dir =~ /\s/;

        my $replace_table = $blog_config->{replace_table};

        my $compress_format = $blog_config->{compress_format} || 'zip';
        my $compress_file = File::Spec->catdir($base_dir, "download-blog-$blog_id");
        $compress_file = qq("$compress_file") if $compress_file =~ /\s/;

        my $transfer = $blog_config->{transfer} || 'ftp';

        my $exclude_dirs = $blog_config->{exclude_dirs};

        my $ftp_server = $blog_config->{ftp_server};
        my $ftp_user = $blog_config->{ftp_user};
        my $ftp_password = $blog_config->{ftp_password};
        my $ftp_remote_dir = $blog_config->{ftp_remote_dir};

        my $sync_dir = $blog_config->{sync_dir};
        $sync_dir = qq("$sync_dir") if $sync_dir =~ /\s/;

        my $custom_actions = $blog_config->{custom_actions} || '';

        my $start = time;

        my @steps = qw/copy replace transfer delete/;
        push @steps, 'custom' if $custom_actions;

        progress("Start batch process of Uploader for a blog '[_1]([_2])'.", $blog->name, $blog->id);

        local $@;
        my $rest = scalar @steps;
        eval {
            while ( my $step = shift @steps ) {
                $rest--;
                if ( $step eq 'copy' ) {
                    step_copy($config, $blog_dir, $temp_dir, $exclude_dirs) || last;
                    progress('Copy blog contents to temporary.');
                }
                elsif ( $step eq 'replace' ) {
                    step_replace($config, $temp_dir, $replace_table) || last;
                    progress('Replace the temporary contents with table.');
                }
                elsif ( $step eq 'compress' ) {
                    if ( $compress_format eq 'zip' ) {
                        step_compress_zip($config, $temp_dir, $compress_file) || last;
                    } else {
                        step_compress_targz($config, $temp_dir, $compress_file) || last;
                    }
                    progress('Compress the temporary contents.');
                }
                elsif ( $step eq 'transfer' ) {
                    if ( $transfer eq 'ftp' ) {
                        my $res = step_transfer_ftp($config, $temp_dir, $ftp_server, $ftp_user, $ftp_password, $ftp_remote_dir);

                        my $msg = 'Transfer and synchronize the temporary contents via FTP.';
                        if ( $res && $res !~ /^1$/ ) {
                            $msg = $res;
                        }
                        progress($msg);
                    }
                    else {
                        step_transfer_local($config, $temp_dir, $sync_dir) || last;
                        progress('Transfer and synchronize the temporary contents to the directory.');
                    }
                }
                elsif ( $step eq 'delete' ) {
                    step_delete($config, $temp_dir) || last;
                    progress('Delete the temporary contents.');
                }
                elsif ( $step eq 'custom' ) {
                    step_custom($config, $custom_actions, $base_dir, $temp_dir, $blog_id) || last;
                    progress('Process custom actions.');
                }
                else {
                    MT::log('[_1] is bad step', $step);
                    last;
                }
            }
        };

        MT::log($@) if $@;
        MT::log( $plugin->translate('An error occurred.') ) if $rest;

        if ( $@ or $rest ) {
            increment_retry($plugin, $blog, $blog_config, $config)
                if $blog_config->{reservation_retried} >= $config->{reservation_retry};
            next;
        }

        next if $rest;
        progress("Finish batch process of Uploader for a blog '[_1]([_2])'.", $blog->name, $blog->id);

        if ( $blog_config->{reservation} ) {
            burndown_reserve($plugin, $blog, $blog_config, $config);
            MT::log(
                $plugin->translate('Finished reserved transfer abount blog: [_1]',
                    $blog->name,
                )
            );
        }
    }
}

sub burndown_reserve {
    my ( $plugin, $blog, $blog_config, $config ) = @_;
    $blog_config->{last_reserve_transfered_epoch} = time;
    $plugin->save_config($blog_config, 'blog:' . $blog->id);
}

sub increment_retry {
    my ( $plugin, $blog, $blog_config, $config ) = @_;
    $blog_config->{reservation_retried} ||= 0;
    $blog_config->{reservation_retried} ++;

    my $res = 1;
    if ( $blog_config->{reservation_retried} > $config->{reservation_retry} ) {
        $blog_config->{reservation_retried} = $config->{reservation_retry};
        $blog_config->{last_reserve_transfered_epoch} = time;
        $res = 0;
    }
    $plugin->save_config($blog_config, 'blog:' . $blog->id);

    $res;
}

sub sync_blogs {
    my $plugin = plugin();

    my @blogs = MT::Blog->load();
    if ( MT->version_number >= 5 ) {
        require MT::Website;
        my @websites = MT::Website->load();
        push @blogs, @websites;
    }

    my @targets;

    foreach ( @blogs ) {
        my $config = $plugin->get_config_hash('blog:' .$_->id);
        push @targets, $_
            if $config->{'do_run-periodical-tasks'}
                || $config->{reservation};
    }

    if ( MT->config->UploaderDebugMode ) {
        my @target_ids = map { $_->id } @targets;
        MT::log( 'sync_targets: ' .join(',', @target_ids) );
    }

    \@targets;
}

sub plugin {
    MT::Plugin::Uploader->instance;
}


sub get_configs {
    my $blog = shift;

    my $plugin = plugin();

    my $blog_id = $blog->id;
    return MT::log( $plugin->translate('blog_id is required') ) unless $blog_id;

    my $mt_path = MT->instance->mt_dir;
    my $plugin_dir = File::Spec->catdir($mt_path, $plugin->envelope);

    my $config = $plugin->get_config_hash();
    my $blog_config = $plugin->get_config_hash('blog:'.$blog_id);

    my $base_dir = $config->{temp_dir} || File::Spec->catdir($plugin_dir, '/tmp/');

    return ($config, $blog_config, $mt_path, $blog_id, $base_dir);
}

sub step_copy {
    my ($config, $blog_dir, $temp_dir, $exclude_dirs) = @_;
    my $result;
    my $script;
    my @exclude_dir_list;
    my $exclude_command;

    step_delete($config, $temp_dir);

    @exclude_dir_list = grep { $_ } map { s/(^\s+)|(\s+$)//g; $_ } split(/\r?\n/, $exclude_dirs);

    my $unique_temp_dir = MT->instance->config('UploaderUniqueTempDir');
    my $copy_with_rsync = MT->instance->config('UploaderCopyWithRsync');
    if ( !$unique_temp_dir && $copy_with_rsync ) {
        my $rsync = MT->instance->config('UploaderRsyncPath') || 'rsync';
        my $excludes = join ' ', map { s/'/\\'/; "--exclude '$_'" } @exclude_dir_list;
        my $options = MT->instance->config('UploaderRsyncCopyOption') || '-auv --delete';
        $script = "$rsync $options '$blog_dir/' '$temp_dir' $excludes 2>&1";
    } else {
        $script = "cp -a '$blog_dir' '$temp_dir' 2>&1";
    }

    MT::log($script) if $MT::DebugMode;
    $result = `$script`;
    MT::log($result) if $MT::DebugMode;

    if ( $? != 0 ) {
        MT::log($result);
        return 0;
    }

    if ( @exclude_dir_list ) {
        foreach ( @exclude_dir_list ) {
            my $del_path = File::Spec->catdir($temp_dir, $_);
            next unless defined $del_path && -e $del_path;
            $script = "rm -rf '$del_path'";
            $result = `$script`;
            if ( $? != 0 ) {
                MT::log($result);
                return 0;
            }
        }
    }

    1;
}

sub step_replace {
    my ($config, $temp_dir, $table) = @_;
    my $result;
    my $script;

    if( $config->{replace_script} ){
        $script = "$config->{replace_script} 2>&1";
        $result = `$script`;
        if ( $? != 0 ) {
            MT::log($result);
            return 0;
        }
    }
    else {
        $table =~ s/\r//g;
        foreach (split(/\n/, $table)) {
            my ($cond, $exec) = split(/:/, $_, 2);
            $cond = "*.$1" if $cond =~ /^\*([^.].*?)$/;
            if($exec) {
                $cond = $cond? "-name '$cond'": '-type f';
                $script = "find '$temp_dir' $cond -print0 | xargs -0 sed -i -e '$exec' 2>&1";
                $result = `$script`;
                if ( $? != 0 && $result !~ /no input files/) {
                    MT::log($result ."temp_dir:$temp_dir, cond:$cond, exec:$exec");
                    return 0;
                }
            }
        }
    }
    1;
}

sub step_compress_zip {
    my ($config, $temp_dir, $compress_file) = @_;
    my $result;
    my @scripts;

    if( $config->{zip_script} ){
        my $script = "$config->{zip_script} 2>&1";
        $result = `$script`;
    }
    else {
        @scripts = ("cd $temp_dir;zip -r $compress_file.zip ./* 2>&1",
            "mv $compress_file.zip $compress_file 2>&1");
        foreach( @scripts ){
            $result = `$_`;
            if ( $? != 0 ) {
                MT::log($result.", command=$_, status=$?");
                return 0;
            }
        }
    }
    1;
}

sub step_compress_targz {
    my ($config, $temp_dir, $compress_file) = @_;
    my $result;
    my @scripts;

    @scripts = (
        "cd $temp_dir;tar cvzf $compress_file.tar.gz ./* 2>&1",
        "mv $compress_file.tar.gz $compress_file 2>&1");
    @scripts = ("$config->{tar_script} 2>&1") if $config->{tar_script} ;

    foreach( @scripts ){
        $result = `$_`;
        if ( $? != 0 ) {
            MT::log($result);
            return 0;
        }
    }

    1;
}

sub step_transfer_ftp {
    my ($config, $temp_dir, $ftp_server, $ftp_user, $ftp_password, $ftp_remote_dir) = @_;
    my $result;
    my $script;
    my $options = $config->{lftp_mirror_options} || '-R -n -e';
    my $lftp_pre_command = $config->{lftp_pre_command} || '';
    $lftp_pre_command .= ' &&' if $lftp_pre_command;

    $script = "lftp -c \"$lftp_pre_command open $ftp_server && user $ftp_user $ftp_password && lcd $temp_dir && mirror $options ./ $ftp_remote_dir \" 2>&1";
    $script = "$config->{ftp_script} 2>&1" if $config->{ftp_script} ;
    MT::log($script) if $MT::DebugMode;

    $@ = undef;

    $result = `$script`;

    if ( $? != 0 ) {
        $result = Encode::is_utf8($result) ? $result : Encode::decode_utf8($result);
        MT::log(translate('Failed to Transfer via FTP. [_1]', $result));
        return translate('Failed to Transfer via FTP. [_1]', $result);
    }
    1;
}

sub step_transfer_local {
    my ($config, $temp_dir, $sync_dir) = @_;
    my $result;
    my $script;

    my $rsync = MT->instance->config('UploaderRsyncPath') || 'rsync';
    $script = "$rsync -auv --delete '$temp_dir/' '$sync_dir' 2>&1";
    if ( $config->{rsync_script} ) {
        $script = "$config->{rsync_script} '$temp_dir/' '$sync_dir' 2>&1";
    }
    MT::log($script) if $MT::DebugMode;

    $result = `$script`;
    if ( $? != 0 ) {
        MT::log($result);
        return 0;
    }

    1;
}

sub step_delete {
    my ($config, $temp_dir) = @_;
    my $result;
    my $script;

    my $unique_temp_dir = MT->instance->config('UploaderUniqueTempDir');
    return 1 if $unique_temp_dir;

    my $copy_with_rsync = MT->instance->config('UploaderCopyWithRsync');
    return 1 if $copy_with_rsync;

    $script = "rm -rf '$temp_dir' 2>&1";
    $script = "$config->{delete_script} 2>&1" if $config->{delete_script} ;

    my @parts;
    @parts = split(/\//, $temp_dir);
    @parts = split(/\\/, $temp_dir) unless @parts;
    if ( @parts < 3 ) {
        MT::log(translate('About to delete shallow directory [_1]', $temp_dir));
        return 0;
    }
    $result = `$script`;
    if ( $? != 0 ) {
        MT::log($result);
        return 0;
    }
    1;
}

sub step_custom {
    my ($config, $custom_actions, $base_dir, $temp_dir, $blog_id) = @_;

    $custom_actions =~ s/\r//g;
    $custom_actions =~ s/\[base_dir\]/$base_dir/g;
    $custom_actions =~ s/\[temp_dir\]/$temp_dir/g;
    $custom_actions =~ s/\[blog_id\]/$blog_id/g;

    my @commands = split(/\n/, $custom_actions);

    foreach (@commands){
        my $result = `$_`;
        if ( $? != 0 ) {
            MT::log($result. ", command=$_, status=$?");
            return 0;
        }
    }

    1;
}

sub progress {
    my $msg = shift;
    return unless MT->config->UploaderDebugMode;
    my $plugin = plugin();
    my $log = MT::Util::log_time() .$plugin->translate($msg, @_);
    MT::log($log);
}

sub translate {
    my $plugin = plugin();
    $plugin->translate(@_);
}

sub translate_json {
    my $app = shift;
    my $trans = $app->plugin->translate(@_);
    $trans = MT::I18N::encode_text($trans, undef, 'utf-8');
    return MT::Util::escape_unicode($trans);
}

sub translate_templatized {
    my $app = shift;
    $app->plugin->translate_templatized(@_);
}

1;
