# uploader.pl

package MT::Plugin::Uploader;
use vars qw( $VERSION $TRIAL_DAYS $CONFIRM_INTERVAL );
$VERSION = "1.401";
$TRIAL_DAYS = 45;
$CONFIRM_INTERVAL = 1209600;

# ライブラリ
use File::Spec;
use MT;
use MT::Util;
use Uploader::Task;
use Uploader::L10N;
use CherryPick::LicenseManager;
use CherryPick::LicenseManager::CMS;

# プラグマ
use strict;

# 基底パッケージ
use base qw(MT::Plugin);

# プラグイン情報とその登録
my $plugin = MT::Plugin::Uploader->new({
  name => 'Uploader',
  key => __PACKAGE__,
  version => $VERSION,
  description => MT->version_number >= 3.3?
    "<MT_TRANS phrase=\"Support to upload blog contents to another server.\">":
    convert_from_utf8('ブログのコンテンツを他のサーバーにアップロードする作業を支援します'),
  doc_link => 'https://www.cherry-pick.jp/mt-plugins/uploader/',
  author_name => "cherry-pick Inc.",
  author_link => 'https://www.cherry-pick.jp/mt-plugins/uploader/',
  l10n_class => 'Uploader::L10N',
  config_template => \&_hdr_config_template,
  settings => new MT::PluginSettings([
      ['temp_dir'],
      ['replace_table'],
      ['transfer'],
      ['ftp_server'],
      ['ftp_user'],
      ['ftp_password'],
      ['ftp_remote_dir'],
      ['lftp_pre_command'],
      ['lftp_mirror_options'],
      ['exclude_dirs'],
      ['sync_dir'],
      ['compress_format'],
      ['copy_script'],
      ['replace_script'],
      ['zip_script'],
      ['tar_script'],
      ['ftp_script'],
      ['rsync_script'],
      ['delete_script'],
      ['custom_actions'],
      ['do_run-periodical-tasks'],
      ['reservation', { Default => 0 }],
      ['reservation_epoch', { Default => 0 }],
      ['reservation_retried', { Default => 0 }],
      ['last_reserve_transfered_epoch', { Default => 0 }],
      ['reservation_retry', { Default => 3, Scope => 'system' }],

      # FTP Proxy対応
      ['http_proxy_user'],
      ['http_proxy_pass'],
      ['http_proxy_host'],
      ['http_proxy_port'],
      ['use_proxy'],
      ['proxy_type'],
  ]),
});

sub init_registry {
  my $plugin = shift;
  $plugin->registry({
    config_settings => {
        'UploaderUniqueTempDir' => { Default => 0 },
        'UploaderUniqueTempDirExpires' => { Default => 3600 },
        'UploaderCopyWithRsync' => { Default => 0 },
        'UploaderRsyncPath' => { Default => 'rsync' },
        'UploaderLftpPath' => { Default => 'lftp' },
        # FTP Proxy対応
        'UploaderEnableProxy' => { Default => 0 },
        'UploaderWarningTransferTimeout' => { Default => 0 },
    },
    permissions => {
        'blog.uploader' => {
            'group'        => 'blog_admin',
            'label'            => 'Transfer and Download(Uploader)',
            'order'            => 2000,
            'permitted_action' => {
                'transfer_contents' => 1,
                'download_contents' => 1,
            },
        },
        'blog.administer_website' => {
            'inherit_from'     => ['blog.administer_blog'],
        },
        'blog.administer_blog' => {
            'inherit_from'     => ['blog.uploader'],
        },
    },
    applications => {
      uploader => {
        handler => 'Uploader::App',
        cgi_path => sub {
          my $admin_path = MT->instance->config->AdminCGIPath;
          my $p = MT->component('uploader/uploader.pl');
          MT::Util::caturl($admin_path, $p->envelope);
        },
        script => sub { 'uploader.cgi' },
      },
      cms => {
        methods => {
          uploader_download => '$Uploader::Uploader::CMS::download',
          uploader_transfer => '$Uploader::Uploader::CMS::transfer',
          uploader_reserve => '$Uploader::Uploader::App::reserve',
          uploader_save_reserve => '$Uploader::Uploader::App::save_reserve',
          uploader_register_license_manager => sub {
            CherryPick::LicenseManager::CMS::handle_register_licensekey(MT::Plugin::Uploader->instance->license_manager, @_);
          },
        },
        menus => {
          'upload' => {
            label => 'Upload',
            order => 760,
          },
          'upload:transfer' => {
            label => 'Transfer',
            mode => 'uploader_transfer',
            order => 100,
            permission => 'administer_blog,uploader',
            view => ['blog', 'website'],
          },
          'upload:reserve' => {
            label => 'Reserve Transfer',
            mode => 'uploader_reserve',
            order => 200,
            permission => 'administer_blog,uploader',
            view => ['blog', 'website'],
            condition => sub {
              my $q = MT->instance->param;
              my $blog_id = $q->param('blog_id');
              my %config;
              # MT->component('Uploader')->load_config(\%config, "blog:$blog_id");

              1;
            },
          },
          'upload:download' => {
            label => 'Download',
            mode => 'uploader_download',
            order => 300,
            permission => 'administer_blog,uploader',
            view => ['blog', 'website'],
          },
        },
        callbacks => {
          'template_param.cfg_plugin' => sub {
            CherryPick::LicenseManager::CMS::handle_template_param_cfg_plugin(MT::Plugin::Uploader->instance->license_manager, @_);
          },
        },
      },
    },
    tasks => {
      'Uploader' => {
        key => 'Uploader',
        label => 'Uploade files',
        frequency => 1,
        code => sub {
          Uploader::Task->upload;
        },
      },
    },
  });
}

sub instance {
  $plugin;
}

sub license_manager {
    my $plugin = shift;

    my $local_active_env_name = $plugin->name .'LocalActivation';
    my $lm = CherryPick::LicenseManager->new({
        plugin          => $plugin,
        register_url    => 'https://mtlmc.cherry-pick.jp/api/licenses/register',
        confirm_url     => 'https://mtlmc.cherry-pick.jp/api/licenses/register',
        confirm_interval => $CONFIRM_INTERVAL,
        trial_days => $TRIAL_DAYS,
        local_activation => MT->config->$local_active_env_name ? 1 : 0,
    }) || return $plugin->error(CherryPick::LicenseManager->errstr);

    $lm;
}

sub is_enable {
  my $lm = $plugin->license_manager
    || return $plugin->errstr;
  return ($lm->is_registered || !$lm->is_expired) ? 1 : 0
}


MT->add_plugin($plugin);

MT->add_plugin_action('blog','uploader.cgi?__mode=download', "Download contents");
MT->add_plugin_action('blog','uploader.cgi?__mode=conftransfer', "Transfer contents");

# 国際化対応(3.2のフォロー)
sub translate {
  my $plugin = shift;

  # 3.3以上であればプラグイン標準の国際化処理を行う
  return $plugin->SUPER::translate(@_) if MT->version_number >= 3.3;

  # 3.2であればJ10Nモジュールを使い、同じ引数を処理する
  unless($plugin->{__j10n_handle}) {
    require Uploader::J10N;
    $plugin->{__j10n_handle} = Uploader::J10N->get_handle('ja');
  }
  my ($format, @args) = @_;
  my $enc = MT->instance->config('PublishCharset');
  my $str;
  if ($enc =~ m/utf-?8/i) {
    $str = $plugin->{__j10n_handle}->maketext($format, @args);
  } else {
    $str = MT::I18N::encode_text($plugin->{__j10n_handle}->maketext($format, map {MT::I18N::encode_text($_, $enc, 'utf-8')} @args), 'utf-8', $enc);
  }
  if(!defined $str) {
    $str = MT->translate(@_);
  }
  $str;
}

# 国際化対応フォロー(3.2のフォロー)
sub translate_templatized {
    my $plugin = shift;

    # 3.3以上であればプラグイン標準の国際化処理を行う
    return $plugin->SUPER::translate_templatized(@_) if MT->version_number >= 3.3;

    # 3.2であればJ10Nモジュールを使い、MT_TRANSタグを処理する
    my($text) = @_;
    $text =~ s!(<MT_TRANS(?:\s+((?:\w+)\s*=\s*(["'])(?:<[^>]+?>|[^\3]+?)+?\3))+?\s*/?>)!
        my($msg, %args) = ($1);
        while ($msg =~ /\b(\w+)\s*=\s*(["'])((?:<[^>]+?>|[^\2])*?)\2/g) {  #"
            $args{$1} = $3;
        }
        $args{params} = '' unless defined $args{params};
        my @p = map MT::Util::decode_html($_),
            split /\s*%%\s*/, $args{params};
        @p = ('') unless @p;
        my $translation = $plugin->translate($args{phrase}, @p);
        $translation =~ s/([\\'])/\\$1/sg if $args{escape};
        $translation;
    !ge;
    $text;
}

# コマンドの有無をチェック
sub check_commands {
  my @commands = qw(rsync lftp mv cp rm find sed xargs tar zip);
  my (@unabled, @enabled);
  foreach (@commands) {
    my $log = `lftp -v`;
    my $which = [split(/\n/, `which --skip-alias "$_"`)]->[0];
    $which = [split(/\n/, `which "$_"`)]->[0] unless $which;
    if(-e $which && -x $which) {
      push @enabled, {command => $_, path => $which};
    } else {
      push @unabled, {command => $_};
    }
  }

  return (\@enabled, \@unabled);
}

# プラグインの管理画面

# プラグインの設定フォームテンプレートハンドラ
sub _hdr_config_template {
  my $plugin = shift;
  my ($param, $scope) = @_;

  my $template;

  # ライセンスキー登録用テンプレート
  my $lm = $plugin->license_manager
      || return $plugin->errstr;
  $param->{email} = $lm->email;
  $param->{licensekey} = $lm->licensekey;
  $param->{is_registered} = $lm->is_registered ? 1 : 0;
  $param->{is_expired} = $lm->is_expired ? 1 : 0;
  $param->{remainder_days} = $lm->remainder_days;

  # システムコンテキストの場合
  if($scope eq 'system') {
    # MTのパス
    my $mt_path = MT->instance->mt_dir;
    my $plugin_dir = File::Spec->catdir($mt_path, $plugin->envelope);

    # デフォルト値設定
    # テンポラリディレクトリ
    $param->{temp_dir} = File::Spec->catdir($plugin_dir, '/tmp/') unless $param->{temp_dir};

    # 警告
    # テンポラリディレクトリが書き込み不可
    $param->{temp_dir_not_writable} = 1 unless -w $param->{temp_dir};

    $template = qq{
<style type="text/css">
<!--
  .good {
    color: #93b06b;
    padding-top : 0px;
    margin-top : 0px;
  }
  .bad {
    padding-top : 0px;
    margin-top : 4px;
    border-left : 1px solid red;
    padding-left : 10px;
    margin-left : 60px;
  }
-->
</style>

<!-- Temporary diectory -->
<mtapp:setting
    id="temp_dir"
    label="<MT_TRANS phrase="Temp directory">"
    hint="<MT_TRANS phrase="Enter the temporary directory path.">"
    show_hint="1">
    <input type="text" name="temp_dir" id="temp_dir" value="<mt:var name="temp_dir" escape="html">" />
    <TMPL_IF NAME=TEMP_DIR_NOT_WRITABLE><p class="bad"><MT_TRANS phrase="This directory is not writable."></p>
</TMPL_IF>
</mtapp:setting>

<!-- Run periodical-tasks -->
<mtapp:setting
    id="reservation-retry"
    label="<MT_TRANS phrase="Retry count">"
    hint="<MT_TRANS phrase="Retry count for reserved transfer.">"
    show_hint="1">
<input type="text" class="text num" name="reservation_retry" id="uploader-reservation-retry" value="<TMPL_VAR NAME=RESERVATION_RETRY ESCAPE=HTML>" />
</mtapp:setting>
};

    # FTP Proxy対応
    $template .= qq{

<!-- Proxy settings -->
<mtapp:setting
    id="use_proxy"
    label="<MT_TRANS phrase="Use Proxy">"
    >
    <p><MT_TRANS phrase="Check the bellow, if you use proxy."></p>
    <input type="checkbox" name="use_proxy" id="use_proxy" value="1"<mt:if name="use_proxy"> checked="checked"</mt:if> />&nbsp;<__trans phrase="Use Proxy" />
</mtapp:setting>
<mtapp:setting
    id="proxy_type"
    label="<MT_TRANS phrase="Proxy Type">"
    >
    <p><MT_TRANS phrase="Select proxy type."></p>
    <input type="radio" name="proxy_type" id="proxy_type_hftp" value="hftp"<mt:if name="proxy_type" eq="hftp"> checked="checked"</mt:if> />&nbsp;<__trans phrase="FTP over HTTP Proxy" /><br />
    <input type="radio" name="proxy_type" id="proxy_type_ftp" value="ftp"<mt:unless name="proxy_type" eq="hftp"> checked="checked"</mt:unless> />&nbsp;<__trans phrase="FTP Proxy" />
</mtapp:setting>
<mtapp:setting
    id="http_proxy_user"
    label="<MT_TRANS phrase="Proxy Username">"
    >
    <p><MT_TRANS phrase="Enter username of Proxy."></p>
    <input type="text" name="http_proxy_user" id="http_proxy_user" value="<mt:var name="http_proxy_user">" />
</mtapp:setting>
<mtapp:setting
    id="http_proxy_pass"
    label="<MT_TRANS phrase="Proxy Password">"
    >
    <p><MT_TRANS phrase="Enter password of Proxy."></p>
    <input type="password" name="http_proxy_pass" id="http_proxy_pass" value="<mt:var name="http_proxy_pass">" />
</mtapp:setting>
<mtapp:setting
    id="http_proxy_host"
    label="<MT_TRANS phrase="Proxy Host">"
    >
    <p><MT_TRANS phrase="Enter host name or IP address of Proxy."></p>
    <input type="text" name="http_proxy_host" id="http_proxy_host" value="<mt:var name="http_proxy_host">" />
</mtapp:setting>
<mtapp:setting
    id="http_proxy_port"
    label="<MT_TRANS phrase="Proxy Port">"
    >
    <p><MT_TRANS phrase="Enter port number of Proxy."></p>
    <input type="text" name="http_proxy_port" id="http_proxy_port" value="<mt:var name="http_proxy_port">" />
</mtapp:setting>

} if MT->instance->config('UploaderEnableProxy');

    # 必要なコマンドの有無を確認する
    my ($enabled, $unabled) = check_commands;
    $param->{unabled_commands} = $unabled;

    $template .= qq{

<mtapp:setting
    id="unix-command"
    label="<__trans phrase="Unix Command:">"
    >
<TMPL_IF NAME=UNABLED_COMMANDS>
<p><MT_TRANS phrase="Some UNIX command is not discovered:"></p>
<TMPL_LOOP NAME=UNABLED_COMMANDS>
<p class="bad"><TMPL_VAR NAME=COMMAND></p>
</TMPL_LOOP>
<TMPL_ELSE>
<p><MT_TRANS phrase="All required UNIX command is exists."></p>
</TMPL_IF>
</mtapp:setting>

};

    $template = $plugin->translate_templatized($template) unless MT->version_number >= 3.3;

      my $lm_tmpl = <<'EOT';
<mt:if name="is_expired">
    <p><__trans phrase="Trial is expired."></p>
<mt:else>
    <mt:unless name="is_registered">
        <mtapp:setting
            id="probation"
            label="<__trans phrase="Probation:">"
            hint="<__trans phrase="Hint">">
            <__trans phrase="Ends in"> <mt:var name="remainder_days"> <__trans phrase="days">
        </mtapp:setting>
    </mt:unless>
</mt:if>
<mtapp:setting
    id="license-key-pref"
    label="<__trans phrase="License key:">"
    hint="<__trans phrase="Hint">">
    <input type="text" name="licensekey" id="license-key" value="<mt:var name="licensekey" escape="html">" />
</mtapp:setting>
<mtapp:setting
    id="email-pref"
    label="<__trans phrase="E-mail:">"
    hint="<__trans phrase="Hint">">
    <input type="text" name="email" id="email" value="<mt:var name="email" escape="html">" />
</mtapp:setting>
<div class="actions-bar settings-actions-bar">
    <button
        mt:mode="save_plugin_config"
        class="save action button primary primary-button"
        onclick="this.form['__mode'].value = 'uploader_register_license_manager'; this.form.submit(); return false;"
        ><__trans phrase="Register license"></button>
</div>
EOT
    my $html = $lm_tmpl;
    $html .= $template unless $param->{is_expired};
    return $html;
  }
  else {
    return $plugin->translate("Expired trial days. Register your licensekey.")
      unless $plugin->is_enable;

    # ブログコンテキスト
    $param->{compress_format} ||= 'zip';
    $param->{compress_format_zip} = 1 if $param->{compress_format} eq 'zip';
    $param->{compress_format_targz} = 1 if $param->{compress_format} eq 'tar.gz';
    $param->{transfer} ||= 'ftp';
    $param->{transfer_ftp} = 1 if $param->{transfer} eq 'ftp';
    $param->{transfer_local} = 1 if $param->{transfer} eq 'local';

    # 設定テンプレートHTML
    $template = qq{
<!-- Replace table -->
<p><MT_TRANS phrase="Set replacement table as 'file type':'sed command' for each line."></p>
<p><MT_TRANS phrase="ex) *.html:s/staging.somewhere/www.somewhere/g"></p>
<div class="setting">
<div class="label"><label for="replace_table"><MT_TRANS phrase="Raplace table">:</label></div>
<div class="field">
<textarea name="replace_table" id="replace_table" style="width: 95%; height: 8ex;"><TMPL_VAR NAME=REPLACE_TABLE ESCAPE=HTML></textarea>
</div>
</div>

<!-- Exclude Directory -->
<p><MT_TRANS phrase="Set exclude directory and files when uploading files. When you want to set more than one directory, please split by line feed."></p>
<p><MT_TRANS phrase="ex)<br />img/<br />blog/img/<br />.htaccess<br />.htpasswd"></p>
<div class="setting">
  <div class="label"><label for="exclude_dirs"><MT_TRANS phrase="Exclude directories and files">:</label></div>
  <div class="field">
    <textarea name="exclude_dirs" id="exclude_dirs" style="width: 95%; height: 8ex;"><TMPL_VAR NAME=EXCLUDE_DIRS ESCAPE=HTML></textarea>
  </div>
</div>

<!-- Custom commands -->
<p><MT_TRANS phrase="Set the commands of custom actions to process before download or transfer contents."></p>
<div class="setting">
<div class="label"><label for="custom_actions"><MT_TRANS phrase="Custom actions">:</label></div>
<div class="field">
<textarea name="custom_actions" id="custom_actions" style="width: 95%; height: 8ex;"><TMPL_VAR NAME=CUSTOM_ACTIONS ESCAPE=HTML></textarea>
</div>
</div>

<!-- Compress format -->
<p><MT_TRANS phrase="Select compress format to download contents."></p>
<div class="setting">
<div class="label"><label for="compress_format"><MT_TRANS phrase="Compress format">:</label></div>
<div class="field">
<input type="radio" name="compress_format" value="zip"<TMPL_IF NAME=COMPRESS_FORMAT_ZIP> checked</TMPL_IF> /> <MT_TRANS phrase="zip format"><br />
<input type="radio" name="compress_format" value="tar.gz"<TMPL_IF NAME=COMPRESS_FORMAT_TARGZ> checked</TMPL_IF> /> <MT_TRANS phrase="tar.gz format">
</div>
</div>

<!--Transfer -->
<p><MT_TRANS phrase="Select method to transfer contents."></p>
<div class="setting">
<div class="label"><label for="transfer"><MT_TRANS phrase="Transfer method">:</label></div>
<div class="field">
<input type="radio" name="transfer" value="ftp"<TMPL_IF NAME=TRANSFER_FTP> checked</TMPL_IF> /> <MT_TRANS phrase="FTP upload"><br />
<input type="radio" name="transfer" value="local"<TMPL_IF NAME=TRANSFER_LOCAL> checked</TMPL_IF> /> <MT_TRANS phrase="Local sync">
</div>
</div>

<!-- FTP settings -->
<p><MT_TRANS phrase="Set FTP connection settings if you selected transfer as FTP upload."><p>
<div class="setting grouped">
<div class="label"><label for="ftp_server"><MT_TRANS phrase="FTP server">:</label></div>
<div class="field">
<input type="text" name="ftp_server" id="ftp_server" value="<TMPL_VAR NAME=FTP_SERVER ESCAPE=HTML>" style="width: 95%" />
</div>
</div>

<div class="setting grouped">
<div class="label"><label for="ftp_user"><MT_TRANS phrase="FTP user">:</label></div>
<div class="field">
<input type="text" name="ftp_user" id="ftp_user" value="<TMPL_VAR NAME=FTP_USER ESCAPE=HTML>" style="width: 95%" />
</div>
</div>

<div class="setting grouped">
<div class="label"><label for="ftp_password"><MT_TRANS phrase="FTP password">:</label></div>
<div class="field">
<input type="password" name="ftp_password" id="ftp_password" value="<TMPL_VAR NAME=FTP_PASSWORD ESCAPE=HTML>" style="width: 95%" />
</div>
</div>

<div class="setting">
<div class="label"><label for="ftp_remote_dir"><MT_TRANS phrase="FTP remote dir">:</label></div>
<div class="field">
<input type="text" name="ftp_remote_dir" id="ftp_remote_dir" value="<TMPL_VAR NAME=FTP_REMOTE_DIR ESCAPE=HTML>" style="width: 95%" />
</div>
</div>

<div class="setting">
<p><span style="color:red">[<MT_TRANS phrase="Advanced">]</span> <MT_TRANS phrase="Enter lftp sub-commands run before open sub command. Set empty if not required."><p>
<div class="label"><label for="lftp_pre_command"><MT_TRANS phrase="lftp pre sub-commands">:</label></div>
<div class="field">
<input type="text" name="lftp_pre_command" id="lftp_pre_command" value="<TMPL_VAR NAME=LFTP_PRE_COMMAND ESCAPE=HTML>" style="width: 50%" />
</div>
</div>

<div class="setting">
<p><span style="color:red">[<MT_TRANS phrase="Advanced">]</span> <MT_TRANS phrase="Set mirror command options for lftp if you selected transfer as local sync. If not set, default options are `-R -n -e`."><p>
<div class="label"><label for="lftp_mirror_options"><MT_TRANS phrase="lftp mirror options">:</label></div>
<div class="field">
<input type="text" name="lftp_mirror_options" id="lftp_mirror_options" value="<TMPL_VAR NAME=LFTP_MIRROR_OPTIONS ESCAPE=HTML>" style="width: 50%" />
</div>
</div>

<!-- Local directory -->
<p><MT_TRANS phrase="Set local directory path to sync if you selected transfer as local sync."><p>
<div class="setting">
<div class="label"><label for="sync_dir"><MT_TRANS phrase="Directory to sync">:</label></div>
<div class="field">
<input type="text" name="sync_dir" id="sync_dir" value="<TMPL_VAR NAME=SYNC_DIR ESCAPE=HTML>" style="width: 95%" />
</div>
</div>

<!-- Rsync Command -->
<p><span style="color:red">[<MT_TRANS phrase="Advanced">]</span> <MT_TRANS phrase="Set rsync command as you like if you selected transfer as local sync. If not set, default command is `rsync -auv --delete`."><p>
<div class="setting">
<div class="label"><label for="rsync_script"><MT_TRANS phrase="Rsync Command">:</label></div>
<div class="field">
<input type="text" name="rsync_script" id="rsync_script" value="<TMPL_VAR NAME=RSYNC_SCRIPT ESCAPE=HTML>" style="width: 50%" />
</div>
</div>

<!-- Run periodical-tasks -->
<p><MT_TRANS phrase="Set sync via run-periodica-tasks or not."><p>
<div class="setting">
<div class="label"><label for="sync_dir"><MT_TRANS phrase="Sync via batch">:</label></div>
<div class="field">
<input type="radio" name="do_run-periodical-tasks" id="do_not_run-periodical-tasks" value="0"<TMPL_UNLESS name="DO_RUN-PERIODICAL-TASKS"> checked="checked"</TMPL_UNLESS> /><MT_TRANS phrase="Not sync">&nbsp;
<input type="radio" name="do_run-periodical-tasks" id="do_run-periodical-tasks" value="1"<TMPL_IF name="do_run-periodical-tasks"> checked="checked"</TMPL_IF> /><MT_TRANS phrase="Sync">
</div>
</div>

<input type="hidden" name="reservation" value="<TMPL_VAR NAME=reservation ESCAPE=HTML>" />
<input type="hidden" name="reservation_epoch" value="<TMPL_VAR NAME=reservation_epoch ESCAPE=HTML>" />
<input type="hidden" name="reservation_retried" value="<TMPL_VAR NAME=reservation_retried ESCAPE=HTML>" />
<input type="hidden" name="last_reserve_transfered_epoch" value="<TMPL_VAR NAME=last_reserve_transfered_epoch ESCAPE=HTML>" />
<input type="hidden" name="reservation_retry" value="<TMPL_VAR NAME=reservation_retry ESCAPE=HTML>" />

};

    $template = $plugin->translate_templatized($template) unless MT->version_number >= 3.3;
    return $template;
  }
}


1;
